#include <ctype.h>
#include <string.h>
#include <iostream.h>
#include "mail.h"

static char basis_64[] =
   "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

static int Base64ToInt(char c)
{
  if (c>='A' && c<='Z') return c-'A';
  else if (c>='a' && c<='z') return c-'a'+26;
  else if (c>='0' && c<='9') return c-'0'+52;
  else if (c=='+') return 62;
  else if (c=='/') return 63;
  else if (c=='=') return 64;
  else return 65;
}
// legge i 4 byte in from e li traduce in 3 byte in to
// torna il numero di byte sensati scritti in to
static int Base64ToTriple(unsigned char* from,unsigned char* to)
{
  int quad[4];
  int i;
  for (i=0;i<4 && (quad[i]=Base64ToInt(from[i]))<64;i++);
  to[0]=(quad[0]<<2)|(quad[1]>>4);
  to[1]=((quad[1]&15)<<4)|(quad[2]>>2);
  to[2]=(quad[2]<<6)|(quad[3]);
  if (i<4 && i>0 && quad[i]!=64) 
    cerr<<"Mi aspettavo un '=' a fine file\n";
  return i-1;
}

static int Base64Filter(istream &in,ostream &out)
{
  int n=1;
  unsigned char from[4];
  unsigned char to[3];
  while (in.good()&& n>0)
    {
      while (isspace(in.peek())) in.ignore();
      if (in.peek()=='-') return 1;
      in.read(from,4);
      n=Base64ToTriple(from,to);
      if (n>=0)
	cout.write(to,n);
    }
  return 1;
}

static int begins(char *s,char *t)
{
  for (;*t!='\0' && *s==*t;t++,s++);
  if (*t=='\0') return 1;
  return 0;
}

int Mail::IsBoundary(char *line)
{
  if (father) return father->IsBoundary(line);
  char *s;
  MailCouple *p;
  p=FindCouple("BOUNDARY");
  if (p)
    {
      return line[0]=='-' && line[1]=='-' && begins(line+2,p->value);
    }
  return 0;
}

int Mail::SkipBoundary(void)
{
  char buf[100];
  do
    {
      in->getline(buf,sizeof(buf));
    }
  while(in->good() && !IsBoundary(buf));
  if (in->good())
    while( isspace(in->peek()))in->get();
}

int Mail::CheckStatus(void)
{
  char *s;
  if (status==intosubmail && son && son->status==end)
    status=bodyread;
  if (status==intosubmail && son && son->status==error)
    {
      cerr<<"submail in error\n";
      status=error;
    }
  if (status==intosubmail && son==NULL)
    {
      cerr<<"Submail is null\n";
      status=error;
    }
  if (status==bodyread)
    {
      s=ValueOf("Content-Type:");
      if (s && !strcmp(s,"MULTIPART/MIXED"))
	{
	status=submail;      
	if (in->eof())
	  status=end;
	}
      else 
	status=end;
      
    }
  if (status==end && father)
    {
      if (father->status==submail || father->status==intosubmail)
	{
	father->status=bodyread;
	father->CheckStatus();
	}
    }
  return status!=error;
}
int Mail::SkipBody(void)
{
  if (!CheckStatus()) return 0;
  if (status==intosubmail)
    son->SkipBody();
  else
    {
      if (status==submail || status==preamble)
	{
	  cerr<<"non puoi leggere il body ora\n";
	  status=error;
	  return 0;
	}
      SkipBoundary();
    }
  status=bodyread;
  return CheckStatus();
}

Mail *Mail::ReadSubMail(void)
{
  char *s;
  if (!CheckStatus()) return NULL;
  if (status==intosubmail||status==body)
    {
      SkipBody();
    }
  if (status==end||status==error)
    {
      return NULL;
    }
  if (status!=submail)
    {
      cerr<<"non puoi leggere un submail ora\n";
      status=error;
      return NULL;
    }
  if (son)
    delete son;
  son=new Mail(this);
  return son;
}
int Mail::ReadBody(ostream &out)
{
  char *s;
  if (!CheckStatus()) return 0;
  if (status==intosubmail)
    son->ReadBody(out);
  else
    {
      if (status!=body)
	{
	  cerr<<"non puoi leggere il body ora\n";
	  status=error;
	  return 0;
	}
      s=ValueOf("Content-Transfer-Encoding:");
      if (s==NULL)
	{
	  char buf[100];
	  do
	    {
	      if (in->peek()=='\n')
		cout<<(char)(in->get());
	      in->get(buf,sizeof(buf));
	      if (!IsBoundary(buf))
		out<<buf;
	      else 
		{
		  while (isspace(in->peek())) in->get();
		  break;
		}
	    }
	  while (in->good());
	  status=bodyread;
	}
      else if (!strcmp(s,"BASE64") || !strcmp(s,"base64"))
	{
	  Base64Filter(*in,out);      
	  SkipBoundary();
	  status=bodyread;
	} 
      else
	{
	  cout<<"Unknown encoding type: "<<s<<"\n";
	  SkipBoundary();
	  status=bodyread;
	}
    }
  return CheckStatus();
}

Mail::~Mail()
{
  MailCouple *p,*q;
  for (p=first;p;p=q)
    {
      q=p->next;
      delete p;
    }
  if (son)
    delete son;
}
MailCouple *Mail::FindCouple(char *name)
{
  MailCouple *p;
  for (p=first;p!=NULL && !!strcmp(name,p->name);p=p->next);
  return p;
}
char *Mail::ValueOf(char *name)
{
  MailCouple *p=FindCouple(name);
  if (p)
    return p->value;
  else
    return NULL;
}
int Mail::ReadPreamble(void)
{
  char c;
  if (status!=preamble)
    {
      cerr<<"non puoi leggere il preambolo ora\n";
      status=error;
      return 0;
    }
  while (in->good())
    {
      c=in->peek();
      if (c=='\n') {in->get();status=body;return CheckStatus();}
      AddCouple(new MailCouple(*in));
    }
  cerr<<"non mi aspettavo la fine del preambolo";
  status=error;
  return 0;
}
			 
void Mail::AddCouple(MailCouple *c)
{
  MailCouple **p;
  for (p=&first;(*p)!=NULL;p=&((*p)->next));
  *p=c;
}


MailCouple::MailCouple(istream &in,int level=0)
{
  char *p;
  char buf[256];
  in.getline(buf,sizeof(buf));
  next=NULL;
  this->level=level;
  for (p=buf;*p!='\0' && !isspace(*p);p++);
  if (*p=='\0')
    {
      name=strdup(buf);
      value=strdup("");
    }
  else
    {
      *p='\0';
      name=strdup(buf);
      for (p++;isspace(*p);p++);
      if (!strcmp(name,"Content-Type:"))
	{
	  char *q;
	  for (q=p;*q!='\0' && *q!=';';q++);
	  if (*q=='\0')
	    value=strdup(p);
	  else
	    {
	      *q='\0';
	      value=strdup(p);
	      *(q++)=';';
	      while (isspace(*q)) q++;
	      if (*q!='\0')
		next=new MailCouple(q,level+1);
	    }
	}
      else
	value=strdup(p);
    }
}

//legge i campi del tipo BOUNDARY="adflkfj"; altro=boh; 
MailCouple::MailCouple (char *s,int level)
{
  char *p;
  next=NULL;
  this->level=level;
  for (p=s;*p!='\0' && *p!='=';p++);
  if (*p=='\0')
    {
      name=strdup(s);
      s=p;
      value=strdup("");
    }
  else
    {
      *p='\0';
      name=strdup(s);
      *(p++)='=';
      while (isspace(*p)) p++;
      s=p;
      for ((*s=='"')?(p=s+1):(p=s);*p!='\0' && (*s=='"'?*p!='"':*p!=';');p++);
      if (*p=='\0')
	{
	  value=strdup(s);
	}
      else if (*s=='"')
	{
	  *p='\0';
	  value=strdup(s+1);
	  *(p++)='"';
	  if (*p==';') p++;
	}
      else 
	{
	  *p='\0';
	  value=strdup(s);
	  *(p++)=';';
	}
      while (isspace(*p)) p++;
      s=p;
      if (*s!='\0')
	next=new MailCouple(s,level);
    }
}

void Mail::Out(ostream &o)
{
  MailCouple *p;
  for (p=first;p!=NULL;p=p->next)
    {
      p->Out(o);
    }
}
void MailCouple::Out(ostream &o)
{
  o<<name<<" "<<value<<"\n";
}

ostream *Answer(char *to=NULL)
{
  if (to==NULL)
    {
      to=ValueOf("Reply-To:");
      if (to==NULL)
	to=ValueOf("From:");
      if (to==NULL)
	{
	  cerr<<"Non so a chi rispondere\n";
	  status=error;
	  return NULL;
	}
    }
  char c,s[80];
  
  new opfstream("|sendmail
}
